@@ -13,6 +13,7 @@ from typing import (
     Any,
     Dict,
     Generic,
+    Iterable,
     Iterator,
     List,
     Mapping,
@@ -178,6 +179,18 @@ class AbstractNeuralModel(ABC, Generic[TRawDatapoint, TTensorizedDatapoint, TNeu
         """
         raise NotImplementedError()
 
+    class _TensorizedDataIter(Iterable):
+        def __init__(self, model, dataset_iterator, return_input_data):
+            self._model = model
+            self._dataset_iterator = dataset_iterator
+            self._return_input_data = return_input_data
+
+        def __iter__(self):
+            yield from (
+                (self._model.tensorize(d), d if self._return_input_data else None)
+                for d in self._dataset_iterator
+            )
+
     @final
     def tensorize_dataset(
         self,
@@ -200,11 +213,10 @@ class AbstractNeuralModel(ABC, Generic[TRawDatapoint, TTensorizedDatapoint, TNeu
         """
         assert self.__metadata_initialized, "Metadata has not been initialized."
         if parallelize:
-            base_iterator = (
-                (self.tensorize(d), d if return_input_data else None) for d in dataset_iterator
-            )
+            base_iterator = self._TensorizedDataIter(self, dataset_iterator, return_input_data)
+
             if use_multiprocessing:
-                for tensorized_sample in BufferedIterator(base_iterator):
+                for tensorized_sample in BufferedIterator(base_iterator, max_queue_size=500):
                     if tensorized_sample[0] is not None:
                         yield tensorized_sample
             else:
@@ -331,7 +343,7 @@ class AbstractNeuralModel(ABC, Generic[TRawDatapoint, TTensorizedDatapoint, TNeu
         assert self.__metadata_initialized, "Metadata has not been initialized."
 
         if shuffle_input:
-            tensorized_data = shuffled_iterator(tensorized_data)
+            tensorized_data = shuffled_iterator(tensorized_data, buffer_size=500)
 
         unfinalized_minibatches = ThreadedIterator(
             self.__iterate_unfinalized_minibatches(
