@@ -31,7 +31,7 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
     """
     A trainer for `AbstractComponent`s. Used mainly for supervised learning.
 
-    Create a `ComponentTrainer` by passing a `AbstractComponent` in the constructor.
+    Create a `ModelTrainer` by passing a `AbstractNeuralModel` in the constructor.
     Invoke `train()` to initiate the training loop. The root `TNeuralModule` should return a scalar loss.
     """
 
@@ -67,34 +67,35 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         """
         self.__model = model
         self.__neural_network: Optional[TNeuralModule] = None
-        self.__checkpoint_location = checkpoint_location
+        self._checkpoint_location = checkpoint_location
 
-        self.__max_num_epochs = max_num_epochs
-        self.__minibatch_size = minibatch_size
+        self._max_num_epochs = max_num_epochs
+        self._minibatch_size = minibatch_size
         if optimizer_creator is None:
-            self.__create_optimizer = lambda p: torch.optim.Adam(p)
+            self._create_optimizer = lambda p: torch.optim.Adam(p)
         else:
-            self.__create_optimizer = optimizer_creator
+            self._create_optimizer = optimizer_creator
 
-        self.__create_scheduler = scheduler_creator
+        self._create_scheduler = scheduler_creator
 
         self.__metadata_finalized_hooks: List[Callable[[ModelType], None]] = []
-        self.__training_start_hooks: List[
+        self._training_start_hooks: List[
             Callable[[ModelType, TNeuralModule, torch.optim.Optimizer], None]
         ] = []
-        self.__train_epoch_end_hooks: List[EndOfEpochHook] = []
-        self.__validation_epoch_end_hooks: List[EndOfEpochHook] = []
-        self.__clip_gradient_norm = clip_gradient_norm
-        self.__enable_amp = enable_amp
+        self._train_epoch_end_hooks: List[EndOfEpochHook] = []
+        self._validation_epoch_end_hooks: List[EndOfEpochHook] = []
+        self._improved_epoch_end_hooks: List[EndOfEpochHook] = []
+        self._clip_gradient_norm = clip_gradient_norm
+        self._enable_amp = enable_amp
 
-        self.__target_metric = target_validation_metric
+        self._target_metric = target_validation_metric
         if target_validation_metric is not None:
-            self.__target_metric_higher_is_better = target_validation_metric_higher_is_better
+            self._target_metric_higher_is_better = target_validation_metric_higher_is_better
         else:
             assert (
                 not target_validation_metric_higher_is_better
             ), "When no explicit metric is passed, the validation loss will be used."
-            self.__target_metric_higher_is_better = False
+            self._target_metric_higher_is_better = False
 
     @property
     def model(self) -> AbstractNeuralModel[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]:
@@ -148,15 +149,13 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         )
 
         self.LOGGER.info("Saving model with finalized metadata.")
-        self.__save_checkpoint()
+        self._save_checkpoint()
 
-    def __save_checkpoint(self) -> None:
-        self.__model.save(self.__checkpoint_location, self.neural_module)
+    def _save_checkpoint(self) -> None:
+        self.__model.save(self._checkpoint_location, self.neural_module)
 
-    def __restore_checkpoint(self, device=None) -> None:
-        _, self.neural_module = self.__model.restore_model(
-            self.__checkpoint_location, device=device
-        )
+    def _restore_checkpoint(self, device=None) -> None:
+        _, self.neural_module = self.__model.restore_model(self._checkpoint_location, device=device)
 
     def register_model_metadata_finalized_hook(self, hook: Callable[[ModelType], None]) -> None:
         self.__metadata_finalized_hooks.append(hook)
@@ -164,13 +163,16 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
     def register_training_start_hook(
         self, hook: Callable[[ModelType, TNeuralModule, torch.optim.Optimizer], None]
     ) -> None:
-        self.__training_start_hooks.append(hook)
+        self._training_start_hooks.append(hook)
 
     def register_train_epoch_end_hook(self, hook: EndOfEpochHook) -> None:
-        self.__train_epoch_end_hooks.append(hook)
+        self._train_epoch_end_hooks.append(hook)
 
     def register_validation_epoch_end_hook(self, hook: EndOfEpochHook) -> None:
-        self.__validation_epoch_end_hooks.append(hook)
+        self._validation_epoch_end_hooks.append(hook)
+
+    def register_epoch_improved_end_hook(self, hook: EndOfEpochHook) -> None:
+        self._improved_epoch_end_hooks.append(hook)
 
     def _run_training(
         self,
@@ -188,50 +190,53 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         start_time = time.time()
         self.neural_module.train()
 
-        scaler = torch.cuda.amp.GradScaler(enabled=self.__enable_amp)
+        scaler = torch.cuda.amp.GradScaler(enabled=self._enable_amp)
         with tqdm(desc="Training", disable=not show_progress_bar, leave=False) as progress_bar:
             for step_idx, (mb_data, raw_samples) in enumerate(
                 self.__model.minibatch_iterator(
                     training_tensors(),
                     device=device,
-                    max_minibatch_size=self.__minibatch_size,
+                    max_minibatch_size=self._minibatch_size,
                     yield_partial_minibatches=False,
                     shuffle_input=shuffle_input,
                     parallelize=parallelize,
                 )
             ):
                 optimizer.zero_grad()
-                with torch.cuda.amp.autocast(enabled=self.__enable_amp):
+                with torch.cuda.amp.autocast(enabled=self._enable_amp):
                     mb_loss = self.neural_module(**mb_data)
-                    if torch.isnan(mb_loss):
-                        raise Exception("Loss has a NaN value.")
 
-                    scaler.scale(mb_loss).backward()
+                scaler.scale(mb_loss).backward()
 
-                    if self.__clip_gradient_norm is not None:
-                        scaler.unscale_(optimizer)
-                        torch.nn.utils.clip_grad_norm_(
-                            self.neural_module.parameters(recurse=True), self.__clip_gradient_norm
-                        )
+                if torch.isnan(mb_loss):
+                    raise Exception("Loss has a NaN value.")
 
-                    scaler.step(optimizer)
-                    scaler.update()
-                    if scheduler is not None:
-                        scheduler.step(epoch_idx=epoch, epoch_step=step_idx)
+                if self._clip_gradient_norm is not None:
+                    scaler.unscale_(optimizer)
+                    torch.nn.utils.clip_grad_norm_(
+                        self.neural_module.parameters(recurse=True), self._clip_gradient_norm
+                    )
+
+                scaler.step(optimizer)
+                scaler.update()
+                if scheduler is not None:
+                    scheduler.step(epoch_idx=epoch, epoch_step=step_idx)
 
                 num_minibatches += 1
                 num_samples += len(raw_samples)
-                mb_loss = float(mb_loss.cpu())
-                sum_epoch_loss += mb_loss
-                if num_minibatches == 1:  # First minibatch
-                    running_avg_loss = mb_loss
-                else:
-                    running_avg_loss = (
-                        exponential_running_average_factor * running_avg_loss
-                        + (1 - exponential_running_average_factor) * mb_loss
-                    )
-                progress_bar.update()
-                progress_bar.set_postfix(Loss=f"{running_avg_loss:.2f}")
+                with torch.no_grad():
+                    sum_epoch_loss += mb_loss
+                if show_progress_bar:
+                    mb_loss = float(mb_loss)
+                    if num_minibatches == 1:  # First minibatch
+                        running_avg_loss = mb_loss
+                    else:
+                        running_avg_loss = (
+                            exponential_running_average_factor * running_avg_loss
+                            + (1 - exponential_running_average_factor) * mb_loss
+                        )
+                    progress_bar.update()
+                    progress_bar.set_postfix(Loss=f"{running_avg_loss:.2f}")
 
         elapsed_time = time.time() - start_time
         self.LOGGER.info(
@@ -242,10 +247,12 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         assert (
             num_minibatches > 0
         ), "No training minibatches were created. The minibatch size may be too large or the training dataset size too small."
-        self.LOGGER.info("Epoch %i: Train Loss %.2f", epoch + 1, sum_epoch_loss / num_minibatches)
+        self.LOGGER.info(
+            "Epoch %i: Train Loss %.2f", epoch + 1, float(sum_epoch_loss) / num_minibatches
+        )
         train_metrics = self.neural_module.report_metrics()
 
-        for epoch_hook in self.__train_epoch_end_hooks:
+        for epoch_hook in self._train_epoch_end_hooks:
             epoch_hook(self.__model, self.neural_module, epoch, train_metrics)
 
         if len(train_metrics) > 0:
@@ -263,18 +270,19 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
             for mb_data, raw_samples in self.__model.minibatch_iterator(
                 validation_tensors(),
                 device=device,
-                max_minibatch_size=self.__minibatch_size,
+                max_minibatch_size=self._minibatch_size,
                 yield_partial_minibatches=True,
                 shuffle_input=False,
                 parallelize=parallelize,
             ):
-                with torch.cuda.amp.autocast(enabled=self.__enable_amp):
+                with torch.cuda.amp.autocast(enabled=self._enable_amp):
                     mb_loss = self.neural_module(**mb_data)
                 num_minibatches += 1
                 num_samples += len(raw_samples)
-                sum_epoch_loss += float(mb_loss.cpu())
-                progress_bar.update()
-                progress_bar.set_postfix(Loss=f"{sum_epoch_loss / num_minibatches:.2f}")
+                sum_epoch_loss += mb_loss
+                if show_progress_bar:
+                    progress_bar.update()
+                    progress_bar.set_postfix(Loss=f"{float(sum_epoch_loss) / num_minibatches:.2f}")
 
         elapsed_time = time.time() - start_time
         assert num_samples > 0, "No validation data was found."
@@ -288,22 +296,22 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         self.LOGGER.info("Epoch %i: Valid Loss %.2f", epoch + 1, validation_loss)
 
         validation_metrics = self.neural_module.report_metrics()
-        for epoch_hook in self.__validation_epoch_end_hooks:
+        for epoch_hook in self._validation_epoch_end_hooks:
             epoch_hook(self.__model, self.neural_module, epoch, validation_metrics)
         if len(validation_metrics) > 0:
             self.LOGGER.info("Validation Metrics: %s", json.dumps(validation_metrics, indent=2))
 
-        if self.__target_metric is not None:
-            target_metric = validation_metrics[self.__target_metric]
+        if self._target_metric is not None:
+            target_metric = validation_metrics[self._target_metric]
         else:
             target_metric = validation_loss
 
-        if self.__target_metric_higher_is_better:
+        if self._target_metric_higher_is_better:
             target_metric_improved = target_metric > best_target_metric
         else:
             target_metric_improved = target_metric < best_target_metric
 
-        return target_metric, target_metric_improved
+        return target_metric, target_metric_improved, validation_metrics
 
     def train(
         self,
@@ -320,6 +328,7 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         device=None,
         store_tensorized_data_in_memory: bool = False,
         shuffle_training_data: bool = True,
+        start_epoch_idx: int = 0,
     ) -> None:
         """
         The training-validation loop for `AbstractNeuralModel`s.
@@ -338,6 +347,7 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         :param device: the target PyTorch device for training
         :param store_tensorized_data_in_memory: store all tensorized data in memory instead of computing them on-line.
         :param shuffle_training_data: shuffle the incoming data from `training_data`.
+        :param start_epoch_idx: the idx of the first epoch in this training loop (used for resuming).
         """
         if initialize_metadata:
             self.load_metadata_and_create_network(training_data, parallelize, show_progress_bar)
@@ -366,27 +376,32 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         self.LOGGER.info("Using `%s` for training." % device)
         self.neural_module.to(device)
 
-        optimizer = self.__create_optimizer(self.neural_module.parameters())
-        scheduler = None if self.__create_scheduler is None else self.__create_scheduler(optimizer)
+        optimizer = self._create_optimizer(self.neural_module.parameters())
+        scheduler = None if self._create_scheduler is None else self._create_scheduler(optimizer)
 
-        for hook in self.__training_start_hooks:
+        for hook in self._training_start_hooks:
             hook(self.__model, self.neural_module, optimizer)
 
-        if self.__target_metric_higher_is_better and self.__target_metric is not None:
+        if self._target_metric_higher_is_better and self._target_metric is not None:
             best_target_metric = -math.inf
         else:
             best_target_metric = math.inf
 
         if validate_on_start:
-            target_metric, improved = self._run_validation(
-                validation_tensors, 0, best_target_metric, device, parallelize, show_progress_bar
+            target_metric, improved, _ = self._run_validation(
+                validation_tensors,
+                start_epoch_idx,
+                best_target_metric,
+                device,
+                parallelize,
+                show_progress_bar,
             )
             assert improved
-            self.LOGGER.info(f"Initial {self.__target_metric or 'Loss'}: {target_metric}")
+            self.LOGGER.info(f"Initial {self._target_metric or 'Loss'}: {target_metric}")
             best_target_metric = target_metric
 
         num_epochs_not_improved: int = 0
-        for epoch in range(self.__max_num_epochs):
+        for epoch in range(start_epoch_idx, self._max_num_epochs):
             self._run_training(
                 training_tensors,
                 epoch,
@@ -399,7 +414,13 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
                 shuffle_training_data,
             )
 
-            target_metric, target_metric_improved = self._run_validation(
+            # Save optimizer and epoch id for scheduler
+            torch.save(
+                {"optimizer_state_dict": optimizer.state_dict(), "epoch": epoch + 1},
+                self._checkpoint_location.with_suffix(".optimizerstate"),
+            )
+
+            target_metric, target_metric_improved, validation_metrics = self._run_validation(
                 validation_tensors,
                 epoch,
                 best_target_metric,
@@ -410,12 +431,16 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
             if target_metric_improved:
                 self.LOGGER.info(
                     f"Best performance so far "
-                    f"({self.__target_metric or 'Loss'}: {target_metric:.3f} from {best_target_metric:.3f}). "
+                    f"({self._target_metric or 'Loss'}: {target_metric:.3f} from {best_target_metric:.3f}). "
                     "Saving model checkpoint."
                 )
                 num_epochs_not_improved = 0
-                self.__save_checkpoint()
+                self._save_checkpoint()
                 best_target_metric = target_metric
+
+                if target_metric_improved:
+                    for epoch_hook in self._improved_epoch_end_hooks:
+                        epoch_hook(self.__model, self.neural_module, epoch, validation_metrics)
             else:
                 num_epochs_not_improved += 1
                 if num_epochs_not_improved > patience:
@@ -425,4 +450,4 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
                     break
 
         # Restore the best model params that were found.
-        self.__restore_checkpoint(device=device)
+        self._restore_checkpoint(device=device)
