@@ -10,11 +10,13 @@ Options:
     --max-num-epochs=<epochs>  The maximum number of epochs to run training for. [default: 100]
     --minibatch-size=<size>    The minibatch size. [default: 300]
     --restore-path=<path>      The path to previous model file for starting from previous checkpoint.
+    --restore-optimizer=<path> The path to previous optimizer and scheduler state.
     --sequential-run           Do not parallelize data loading. Makes debugging easier.
     --quiet                    Do not show progress bar.
     -h --help                  Show this screen.
     --debug                    Enable debug routines. [default: False]
 """
+import logging
 import random
 import torch
 from docopt import docopt
@@ -152,16 +154,30 @@ def run(arguments):
     restore_path = arguments.get("--restore-path", None)
     if restore_path:
         initialize_metadata = False
-        model, nn = Graph2Class.restore_model(Path(restore_path))
+        model, nn = Graph2Class.restore_model(Path(restore_path), device="cpu")
     elif arguments["--aml"] and model_path.exists():
         initialize_metadata = False
-        model, nn = Graph2Class.restore_model(model_path)
+        model, nn = Graph2Class.restore_model(model_path, device="cpu")
     else:
         nn = None
         model = create_graph2class_gnn_model()
 
-    def create_optimizer(parameters):
-        return torch.optim.Adam(parameters, lr=0.00025)
+    if arguments["--restore-optimizer"]:
+        opt_state = torch.load(arguments["--restore-optimizer"])
+        current_epoch_idx = opt_state["epoch"]
+
+        def create_optimizer(parameters):
+            opt = torch.optim.Adam(parameters, lr=0.00025)
+            logging.info("Restoring optimizer state from `%s`.", arguments["--restore-optimizer"])
+
+            opt.load_state_dict(opt_state["optimizer_state_dict"])
+            return opt
+
+    else:
+        current_epoch_idx = 0
+
+        def create_optimizer(parameters):
+            return torch.optim.Adam(parameters, lr=0.00025)
 
     trainer = ModelTrainer(
         model,
@@ -192,6 +208,7 @@ def run(arguments):
         parallelize=not arguments["--sequential-run"],
         patience=10,
         store_tensorized_data_in_memory=True,
+        start_epoch_idx=current_epoch_idx,
     )
 
     test_data_path = RichPath.create(arguments["TEST_DATA_PATH"], azure_info_path)
@@ -207,6 +224,9 @@ def run(arguments):
         aml_ctx.log("Test Accuracy", acc)
         aml_ctx.upload_file(name="model.pkl.gz", path_or_stream=str(model_path))
         aml_ctx.upload_file(name="full.log", path_or_stream=log_path)
+        aml_ctx.upload_file(
+            name="optimizer_state.pt", path_or_stream=str(model_path.with_suffix(".optimizerstate"))
+        )
 
 
 if __name__ == "__main__":
